module Main exposing (main)

import Browser
import Element
import Element.Background
import Element.Input
import Puzzles.Day1 as Day1
import Puzzles.Day2 as Day2
import Task
import Tools.Colors as C
import Widget
import Widget.Customize
import Widget.Material as Material
import Widget.Material.Typography as Typography


main : Program () Model Msg
main =
    Browser.document
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }


type alias Model =
    { view : Window
    , input : String
    , output1 : Calculating
    , output2 : Calculating
    }


type Calculating
    = Outcome String
    | Calculating
    | InvalidInput String
    | NoInput


type Window
    = Home
    | Day Int


type alias Puzzle =
    String -> Result String String


getFunctions : Window -> ( Puzzle, Puzzle )
getFunctions window =
    case window of
        Home ->
            ( always <| Err "There is no puzzle for this page!"
            , always <| Err "There is no puzzle for this page!"
            )

        Day 1 ->
            ( Day1.puzzle1, Day1.puzzle2 )

        Day 2 ->
            ( Day2.puzzle1, Day2.puzzle2 )

        _ ->
            ( always <| Err "This puzzle has no implementation yet!"
            , always <| Err "This puzzle has no implementation yet!"
            )


init : () -> ( Model, Cmd Msg )
init () =
    ( { view = Home
      , input = ""
      , output1 = NoInput
      , output2 = NoInput
      }
    , Cmd.none
    )


type Msg
    = ChangeWindow Window
    | OnInput String
    | DoNothing
    | Calculate1 Window (Result String String)
    | Calculate2 Window (Result String String)


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        ChangeWindow window ->
            ( { view = window
              , input = ""
              , output1 = NoInput
              , output2 = NoInput
              }
            , Cmd.none
            )

        OnInput input ->
            ( { model
                | input = input
                , output1 = Calculating
                , output2 = Calculating
              }
            , if input == model.input then
                Cmd.none

              else
                Cmd.batch
                    [ Task.succeed input
                        |> Task.map (getFunctions model.view |> Tuple.first)
                        |> Task.perform (Calculate1 model.view)
                    , Task.succeed input
                        |> Task.map (getFunctions model.view |> Tuple.second)
                        |> Task.perform (Calculate2 model.view)
                    ]
            )

        DoNothing ->
            ( model, Cmd.none )

        Calculate1 w (Ok s) ->
            if w /= model.view then
                ( model, Cmd.none )

            else
                ( { model | output1 = Outcome s }
                , Cmd.none
                )

        Calculate1 w (Err s) ->
            if w /= model.view then
                ( model, Cmd.none )

            else
                ( { model | output1 = InvalidInput s }
                , Cmd.none
                )

        Calculate2 w (Ok s) ->
            if w /= model.view then
                ( model, Cmd.none )

            else
                ( { model | output2 = Outcome s }
                , Cmd.none
                )

        Calculate2 w (Err s) ->
            if w /= model.view then
                ( model, Cmd.none )

            else
                ( { model | output2 = InvalidInput s }
                , Cmd.none
                )


subscriptions : Model -> Sub Msg
subscriptions model =
    Sub.none


view : Model -> Browser.Document Msg
view model =
    { title = "Document Title"
    , body =
        [ Widget.menuBar
            (Material.menuBar palette)
            { title =
                "aoc2023"
                    |> Element.text
                    |> Element.el Typography.h3
            , deviceClass =
                Element.Desktop
            , openLeftSheet = Nothing
            , openRightSheet = Nothing
            , openTopSheet = Nothing
            , primaryActions = []
            , search = Nothing
            }
        , Element.row [ Element.height Element.fill, Element.width Element.fill ]
            [ Element.el
                [ Element.fill
                    |> Element.maximum 200
                    |> Element.width
                , Element.alignTop
                ]
                (List.map
                    (\( window, text ) ->
                        Widget.fullBleedItem
                            (Material.fullBleedItem palette)
                            { text = text
                            , onPress = Just (ChangeWindow window)
                            , icon = always Element.none
                            }
                    )
                    [ ( Home, "Home" )
                    , ( Day 1, "Day 1" )
                    , ( Day 2, "Day 2" )
                    , ( Day 3, "Day 3" )
                    ]
                    |> Widget.itemList Material.column
                )
            , Element.column
                [ Element.height Element.fill
                , Element.width Element.fill
                , C.background (picker C.Dark C.White)
                , Element.padding 20
                , Element.spacing 20
                ]
                (case model.view of
                    Home ->
                        [ "Advent of Code 2023"
                            |> Element.text
                            |> Element.el Typography.h3
                        ]

                    Day i ->
                        [ [ "Using the following input for day "
                                ++ String.fromInt i
                                ++ ": "
                                |> Element.text
                          , Element.Input.multiline
                                [ Element.fill
                                    |> Element.maximum 500
                                    |> Element.height
                                , Element.scrollbarX
                                ]
                                { onChange = OnInput
                                , text = model.input
                                , placeholder =
                                    "Insert puzzle input here..."
                                        |> Element.text
                                        |> Element.Input.placeholder []
                                        |> Just
                                , label = Element.Input.labelHidden "input"
                                , spellcheck = False
                                }
                          ]
                        , case model.output1 of
                            Outcome s ->
                                [ "Part 1 answer:"
                                    |> Element.text
                                , Element.Input.multiline
                                    [ Element.width Element.fill ]
                                    { onChange = always DoNothing
                                    , text = s
                                    , placeholder = Nothing
                                    , label = Element.Input.labelHidden "answer-1"
                                    , spellcheck = False
                                    }
                                ]

                            Calculating ->
                                [ "Calculating part 1..."
                                    |> Element.text
                                ]

                            InvalidInput s ->
                                [ "INVALID INPUT FOR PART 1"
                                    |> Element.text
                                , Element.Input.multiline
                                    [ Element.width Element.fill ]
                                    { onChange = always DoNothing
                                    , text = s
                                    , placeholder = Nothing
                                    , label = Element.Input.labelHidden "answer-1"
                                    , spellcheck = False
                                    }
                                ]

                            NoInput ->
                                List.singleton Element.none
                        , case model.output2 of
                            Outcome s ->
                                [ "Part 2 answer:"
                                    |> Element.text
                                , Element.Input.multiline
                                    [ Element.width Element.fill ]
                                    { onChange = always DoNothing
                                    , text = s
                                    , placeholder = Nothing
                                    , label = Element.Input.labelHidden "answer-2"
                                    , spellcheck = False
                                    }
                                ]

                            Calculating ->
                                [ "Calculating part 2..."
                                    |> Element.text
                                ]

                            InvalidInput s ->
                                [ "INVALID INPUT FOR PART 2"
                                    |> Element.text
                                , Element.Input.multiline
                                    [ Element.width Element.fill ]
                                    { onChange = always DoNothing
                                    , text = s
                                    , placeholder = Nothing
                                    , label = Element.Input.labelHidden "answer-2"
                                    , spellcheck = False
                                    }
                                ]

                            NoInput ->
                                List.singleton Element.none
                        ]
                            |> List.concat
                )
            ]
        ]
            |> Element.column [ Element.width Element.fill, Element.height Element.fill ]
            |> Element.layout [ Element.height Element.fill ]
            |> List.singleton
    }


picker : C.Picker
picker =
    C.get C.Trichromatic C.LightMode


palette : Material.Palette
palette =
    C.defaultPalette picker
