module Puzzles.Day2 exposing (puzzle1, puzzle2)

import Parser as P exposing ((|.), (|=), Parser)


type alias Game =
    { number : Int, games : List RGB }


type alias RGB =
    { red : Int, green : Int, blue : Int }


maxAllowed : RGB
maxAllowed =
    { red = 12, green = 13, blue = 14 }


puzzle1 : String -> Result String String
puzzle1 input =
    case P.run parser input of
        Err a ->
            Debug.log "Output" a
                |> always (Err "Invalid input")

        Ok g ->
            g
                |> List.map
                    (\{ number, games } ->
                        let
                            isLegal : Bool
                            isLegal =
                                List.all
                                    (\rgb ->
                                        List.all
                                            (\f -> f rgb <= f maxAllowed)
                                            [ .red, .green, .blue ]
                                    )
                                    games
                        in
                        if isLegal then
                            number

                        else
                            0
                    )
                |> List.sum
                |> String.fromInt
                |> Ok


puzzle2 : String -> Result String String
puzzle2 input =
    case P.run parser input of
        Err a ->
            Debug.log "Output" a
                |> always (Err "Invalid input")

        -- 179502 too high
        Ok g ->
            g
                |> List.map
                    (\{ games } ->
                        let
                            getMaximumFor : (RGB -> Int) -> Int
                            getMaximumFor f =
                                games
                                    |> List.map f
                                    |> List.maximum
                                    |> Maybe.withDefault 0
                        in
                        [ .red, .green, .blue ]
                            |> List.map getMaximumFor
                            |> List.product
                    )
                |> List.sum
                |> String.fromInt
                |> Ok


parser : Parser (List Game)
parser =
    andOneMore
        { separator = "\n"
        , item = gameParser
        }


gameParser : Parser Game
gameParser =
    P.succeed Game
        |. P.keyword "Game"
        |. P.spaces
        |= P.int
        |. P.symbol ":"
        |. P.spaces
        |= rgbLines


rgbLines : Parser (List RGB)
rgbLines =
    andOneMore
        { separator = ";"
        , item = rgbParser
        }


rgbParser : Parser RGB
rgbParser =
    P.succeed (List.foldl (<|) (RGB 0 0 0))
        |= andOneMore
            { separator = ","
            , item = rgbStmt
            }


rgbStmt : Parser (RGB -> RGB)
rgbStmt =
    P.succeed (|>)
        |= P.int
        |. P.spaces
        |= P.oneOf
            [ P.succeed (\i rgb -> { rgb | red = i })
                |. P.keyword "red"
            , P.succeed (\i rgb -> { rgb | blue = i })
                |. P.keyword "blue"
            , P.succeed (\i rgb -> { rgb | green = i })
                |. P.keyword "green"
            ]


andOneMore : { separator : String, item : Parser a } -> Parser (List a)
andOneMore { separator, item } =
    P.loop []
        (\xs ->
            P.succeed (\x state -> state (x :: xs))
                |= item
                |. onlySpaces
                |= P.oneOf
                    [ P.succeed P.Loop
                        |. P.token separator
                    , P.succeed P.Done
                    ]
                |. onlySpaces
        )


onlySpaces : Parser ()
onlySpaces =
    P.chompWhile (\c -> c == ' ')
